import 'package:flutter/material.dart';
import 'package:flutter_mvp/presenter/i_presenter.dart';
import 'package:flutter_mvp/view/i_view.dart';

/// @desc  基础 widget,关联 Presenter，且与生命周期关联
/// @time 2019-04-22 11:08 am
/// @author cheny
///
abstract class AbstractView extends StatefulWidget {
  AbstractView({Key key}) : super(key: key);
}

// Register the RouteObserver as a navigation observer.
final RouteObserver<PageRoute> routeObserver = RouteObserver<PageRoute>();

abstract class AbstractViewState<T extends IPresenter, V extends AbstractView>
    extends State<V> with WidgetsBindingObserver, RouteAware implements IView {
  T presenter;

  bool _isVisibleToUser =
      false; // 是否对用户可见，表明在onResume-onPause调用过程中。也用来表示当前页面是否在栈顶。

  // onResume、onPause方法事件是否是由系统触发的。也就是AppLifecycleState.resumed和AppLifecycleState.paused中触发。
  // 在从paused到resumed这个过程中，_systemDispatched的值为true。
  bool _systemDispatched = false;

  bool _didPush = false; // 表示是否调用了didPush方法。第一次进入会调用
  bool _didPopNext = false; // 表示是否调用了didPopNext方法。进入别的页面，再次返回时会调用。
  bool _didPop = false; // 表示是否调用了didPop方法。从Navigator中pop时会调用。

  // 一级页面进入二页面再返回，此时回调正常。再次进入二级页面时，不会调用一级页面的onPause方法。
  // 同理，再次返回时也不会调用onResume方法，所以这里引入了一个计数器。
  int _didPopNextCount = 0;

  @override
  void initState() {
    super.initState();
    _onMockCreate();
    _onMockResume();
    WidgetsBinding.instance.addObserver(this);

    presenter = createPresenter();
    if (presenter != null) {
      presenter.attachView(this);
    }
  }

  @override
  void deactivate() {
    super.deactivate();
    if (_didPop || _didPush || _didPopNextCount == 1) {
      _onMockPause();
    } else if (_didPopNext) {
      _onMockResume();
    }
    _resetValues();
  }

  /// 重置这三个变量
  void _resetValues() {
    _didPop = false;
    _didPopNext = false;
    _didPush = false;
    _didPopNextCount--;
  }

  T createPresenter();

  T getPresenter() {
    return presenter;
  }

  @override
  void dispose() {
    _onMockDestroy();
    WidgetsBinding.instance.removeObserver(this);
    routeObserver.unsubscribe(this);
    super.dispose();
    if (presenter != null) {
      presenter.detachView();
      presenter = null;
    }
  }

  @override
  void didChangeAppLifecycleState(AppLifecycleState state) {
    switch (state) {
      case AppLifecycleState.resumed:
        onForeground();
//        _onMockResume();
        _systemDispatched = false;
        break;
      case AppLifecycleState.inactive:
        break;
      case AppLifecycleState.paused:
        onBackground();
        _systemDispatched = true;
//        _onMockPause();
        break;
//      case AppLifecycleState.detached://flutter 1.12变化
//        break;
      default:
        break;
    }
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    routeObserver.subscribe(this, ModalRoute.of(context));
  }

  @override
  void didPush() {
    // Route was pushed onto navigator and is now topmost route.
    _didPush = true;
  }

  @override
  void didPushNext() {
    // Route was pushed onto navigator and is now topmost route.
//    _didPush = true;
  }

  @override
  void didPopNext() {
    // Covering route was popped off the navigator.
    _didPopNext = true;
    _didPopNextCount = 2;
  }

  @override
  void didPop() {
    _didPop = true;
  }

  /// 页面初始化调用，只会调用一次。参照Android端Activity#onCreate方法。
  void _onMockCreate() {
    onCreate();
  }

  /// 页面由不可见变为可见时调用。
  @mustCallSuper
  void _onMockResume() {
    // 满足这两个条件时才会执行：_systemDispatched和_isVisibleToUser均为true，_systemDispatched和_isVisibleToUser均为false。
    if (_systemDispatched && _isVisibleToUser ||
        !_systemDispatched && !_isVisibleToUser) {
      _isVisibleToUser = true;
      onResume();
    }
  }

  /// 页面有可见变为不可见时调用。
  @mustCallSuper
  void _onMockPause() {
    if (!_isVisibleToUser) {
      return;
    }
    // 这一步很关键，在系统传递paused事件过来时，_systemDispatched的值为true，此时不会将_isVisibleToUser的值置为false，是为了跟别的页面区分。
    if (!_systemDispatched) {
      _isVisibleToUser = false;
    }
    onPause();
  }

  /// 页面销毁时调用。
  /// 这里需要说明的时，不管是纯flutter应用还是混合开发的应用，在根flutter页面返回时，都不会调用它的dispose方法。这就需要我们自己
  void _onMockDestroy() {
    onDestroy();
  }

  /// 用于让子类去实现的初始化方法
  void onCreate() {}

  /// 用于让子类去实现的不可见变为可见时的方法
  void onResume() {}

  /// 用于让子类去实现的可见变为不可见时调用的方法。
  void onPause() {}

  /// 用于让子类去实现的销毁方法。
  void onDestroy() {}

  ///从后台回到前吧，子类重写
  void onForeground() {}

  ///从前台压到后台，子类可重写
  void onBackground() {}
}
